@@ -1,4 +1,4 @@
-# Copyright 2023 The dataclass_array Authors.
+# Copyright 2024 The dataclass_array Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -271,7 +271,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
     # Register the tree_map here instead of `__init_subclass__` as `jax` may
     # not have been imported yet during import.
     if enp.lazy.has_jax and not cls._dca_jax_tree_registered:  # pylint: disable=protected-access
-      enp.lazy.jax.tree_util.register_pytree_node_class(cls)
+      enp.lazy.jax.tree_util.register_pytree_with_keys_class(cls)
       cls._dca_jax_tree_registered = True  # pylint: disable=protected-access
 
     if enp.lazy.has_torch and not cls._dca_torch_tree_registered:  # pylint: disable=protected-access
@@ -279,7 +279,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
       # as backend: https://github.com/pytorch/pytorch/issues/65761
       enp.lazy.torch.utils._pytree._register_pytree_node(  # pylint: disable=protected-access
           cls,
-          flatten_fn=lambda a: a.tree_flatten(),
+          flatten_fn=lambda a: a._tree_flatten(),  # pylint: disable=protected-access
           unflatten_fn=lambda vals, ctx: cls.tree_unflatten(ctx, vals),
       )
       cls._dca_torch_tree_registered = True  # pylint: disable=protected-access
@@ -295,7 +295,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
 
     # Validate the batch shape is consistent
     # However, we need to be careful that `_ArrayField` never uses
-    # `@epy.cached_property`
+    # `@functools.cached_property`
     shape = self._broadcast_shape_inplace()
 
     # TODO(epot): When to validate (`field.validate()`)
@@ -348,7 +348,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
       The dataclass array with the new shape
     """
     if isinstance(shape, str):  # Einops support
-      return self._map_field(
+      return self._map_field(  # pylint: disable=protected-access
           array_fn=lambda f: einops.rearrange(  # pylint: disable=g-long-lambda
               f.value,
               np_utils.to_absolute_einops(shape, nlastdim=len(f.inner_shape)),
@@ -365,7 +365,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
       def _reshape(f: _ArrayField):
         return f.value.reshape(shape + f.inner_shape)
 
-      return self._map_field(array_fn=_reshape, dc_fn=_reshape)
+      return self._map_field(array_fn=_reshape, dc_fn=_reshape)  # pylint: disable=protected-access
 
   def flatten(self: _DcT) -> _DcT:
     """Flatten the batch shape."""
@@ -373,7 +373,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
 
   def broadcast_to(self: _DcT, shape: Shape) -> _DcT:
     """Broadcast the batch shape."""
-    return self._map_field(
+    return self._map_field(  # pylint: disable=protected-access
         array_fn=lambda f: f.broadcast_to(shape),
         dc_fn=lambda f: f.broadcast_to(shape),
     )
@@ -456,7 +456,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
       fn: Callable[[Array['*din']], Array['*dout']],
   ) -> _DcT:
     """Apply a transformation on all arrays from the fields."""
-    return self._map_field(
+    return self._map_field(  # pylint: disable=protected-access
         array_fn=lambda f: fn(f.value),
         dc_fn=lambda f: f.value.map_field(fn),
     )
@@ -530,7 +530,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
       array_fn = lambda f: xnp.asarray(f.value)
 
     # Update all childs
-    new_self = self._map_field(
+    new_self = self._map_field(  # pylint: disable=protected-access
         array_fn=array_fn,
         dc_fn=lambda f: f.value.as_xnp(xnp),
     )
@@ -581,9 +581,30 @@ class DataclassArray(metaclass=MetaDataclassArray):
 
   # ====== Internal ======
 
-  @epy.cached_property
+  @functools.cached_property
+  def _all_fields_empty(self) -> bool:
+    """Returns True if the `dataclass_array` is invalid."""
+    if not self._array_fields:  # All fields are `None` / `object`
+      # No fields have been defined.
+      # This can be the case internally by jax which apply some
+      # `tree_map(lambda x: sentinel)`.
+      return True
+
+    # `tf.nest` sometimes replace values by dummy `.` inside
+    # `assert_same_structure`
+    if enp.lazy.has_tf:
+      # pylint: disable=g-direct-tensorflow-import,g-import-not-at-top
+      from tensorflow.python.util import nest_util  # pytype: disable=import-error
+      # pylint: enable=g-direct-tensorflow-import,g-import-not-at-top
+
+      if any(f.value is nest_util._DOT for f in self._array_fields):  # pylint: disable=protected-access,not-an-iterable
+        return True
+    return False
+
+  @functools.cached_property
   def _all_array_fields(self) -> dict[str, _ArrayField]:
     """All array fields, including `None` values."""
+    assert self._dca_fields_metadata is not None
     return {  # pylint: disable=g-complex-comprehension
         name: _ArrayField(
             name=name,
@@ -593,7 +614,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
         for name, field_metadata in self._dca_fields_metadata.items()  # pylint: disable=protected-access
     }
 
-  @epy.cached_property
+  @functools.cached_property
   def _array_fields(self) -> list[_ArrayField]:
     """All active array fields (non-None), including static ones."""
     # Filter `None` values
@@ -603,7 +624,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
 
   def _cast_xnp_dtype_inplace(self) -> Optional[enp.NpModule]:
     """Validate `xnp` are consistent and cast `np` -> `xnp` in-place."""
-    if not self._array_fields:  # All fields are `None` / `object`
+    if self._all_fields_empty:  # pylint: disable=using-constant-test
       return None
 
     # Validate the dtype
@@ -627,12 +648,19 @@ class DataclassArray(metaclass=MetaDataclassArray):
 
     def _cast_field(f: _ArrayField) -> None:
       try:
-        new_value = np_utils.asarray(
-            f.value,
-            xnp=xnp,
-            dtype=f.dtype,
-            cast_dtype=self.__dca_params__.cast_dtype,
-        )
+        # Supports for TensorSpec (e.g. in `tf.function` signature)
+        if enp.lazy.is_tf_xnp(xnp) and isinstance(
+            f.value, enp.lazy.tf.TensorSpec
+        ):
+          # TODO(epot): Actually check the dtype
+          new_value = f.value
+        else:
+          new_value = np_utils.asarray(
+              f.value,
+              xnp=xnp,
+              dtype=f.dtype,
+              cast_dtype=self.__dca_params__.cast_dtype,
+          )
         self._setattr(f.name, new_value)
         # After the field has been set, we validate the shape
         f.assert_shape()
@@ -648,9 +676,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
 
   def _broadcast_shape_inplace(self) -> Optional[Shape]:
     """Validate the shapes are consistent and broadcast values in-place."""
-    if not self._array_fields:  # No fields have been defined.
-      # This can be the case internally by jax which apply some
-      # `tree_map(lambda x: sentinel)`.
+    if self._all_fields_empty:  # pylint: disable=using-constant-test
       return None
 
     # First collect all shapes and compute the final shape.
@@ -735,17 +761,26 @@ class DataclassArray(metaclass=MetaDataclassArray):
       else:
         return array_fn(f)
 
-    new_values = {f.name: _apply_field_dn(f) for f in self._array_fields}  # pylint: disable=not-an-iterable
+    new_values = {f.name: _apply_field_dn(f) for f in self._array_fields}  # pylint: disable=not-an-iterable,protected-access
     # For performance, do not call replace to save the constructor call
     if not _inplace:
       return self.replace(**new_values)
     else:
       return self
 
-  def tree_flatten(self) -> tuple[list[DcOrArray], _TreeMetadata]:
+  def tree_flatten_with_keys(
+      self,
+  ) -> tuple[
+      tuple[tuple[enp.lazy.jax.tree_utils.BuiltInKeyEntry, DcOrArray], ...],
+      _TreeMetadata,
+  ]:
     """`jax.tree_utils` support."""
+    tree_util = enp.lazy.jax.tree_util
     # We flatten all values (and not just the non-None ones)
-    array_field_values = [f.value for f in self._all_array_fields.values()]
+    array_field_values = tuple(
+        (tree_util.GetAttrKey(k), f.value)
+        for k, f in self._all_array_fields.items()
+    )
     metadata = _TreeMetadata(
         array_field_names=list(self._all_array_fields.keys()),
         non_array_field_kwargs={
@@ -792,6 +827,23 @@ class DataclassArray(metaclass=MetaDataclassArray):
         self._setattr(k, v)  # pylint: disable=protected-access
     return self
 
+  def _tree_flatten(self) -> tuple[tuple[DcOrArray, ...], _TreeMetadata]:
+    components, metadata = self.tree_flatten_with_keys()
+    components = tuple(v for _, v in components)
+    return components, metadata
+
+  def __tf_flatten__(self) -> tuple[_TreeMetadata, tuple[DcOrArray, ...]]:
+    components, metadata = self._tree_flatten()
+    return metadata, components
+
+  @classmethod
+  def __tf_unflatten__(
+      cls: Type[_DcT],
+      metadata: _TreeMetadata,
+      components: list[DcOrArray],
+  ) -> _DcT:
+    return cls.tree_unflatten(metadata, components)
+
   def _setattr(self, name: str, value: Any) -> None:
     """Like setattr, but support `frozen` dataclasses."""
     object.__setattr__(self, name, value)
@@ -855,7 +907,7 @@ def _init_cls(self: DataclassArray) -> None:
     # DataclassArray without any array fields
     # Hack: To support `.xnp`, `.shape`, we add a dummy empty field which
     # is propagated by the various ops.
-    dca_fields_metadata[_DUMMY_ARRAY_FIELD] = _ArrayFieldMetadata(
+    dca_fields_metadata[_DUMMY_ARRAY_FIELD] = _ArrayFieldMetadata(  # pytype: disable=wrong-arg-types
         inner_shape_non_static=(),
         dtype=np.float32,
     )
@@ -971,12 +1023,12 @@ class _ArrayFieldMetadata:
   Attributes:
     inner_shape_non_static: Inner shape. Can contain non-static dims (e.g.
       `(None, 3)`)
-    dtype: Type of the array. Can be `array_types.dtypes.DType` or
+    dtype: Type of the array. Can be `enp.dtypes.DType` or
       `dca.DataclassArray` for nested arrays.
   """
 
   inner_shape_non_static: DynamicShape
-  dtype: Union[array_types.dtypes.DType, Type[DataclassArray]]
+  dtype: Union[enp.dtypes.DType, Type[DataclassArray]]
 
   def __post_init__(self):
     """Normalizing/validating the shape/dtype."""
@@ -985,7 +1037,7 @@ class _ArrayFieldMetadata:
 
     # Validate/normalize the dtype
     if not self.is_dataclass:
-      self.dtype = array_types.dtypes.DType.from_value(self.dtype)
+      self.dtype = enp.dtypes.DType.from_value(self.dtype)
       # TODO(epot): Filter invalid dtypes, like `str` ?
 
   def to_dict(self) -> dict[str, Any]:
@@ -1033,7 +1085,7 @@ class _ArrayField(_ArrayFieldMetadata, Generic[DcOrArrayT]):
     # empty shapes to True `bool(shape) == True` when `shape=()`
     return tuple(self.value.shape)
 
-  @epy.cached_property
+  @functools.cached_property
   def inner_shape(self) -> Shape:
     """Returns the the static shape resolved for the current value."""
     if not self.inner_shape_non_static:
@@ -1072,6 +1124,9 @@ class _ArrayField(_ArrayFieldMetadata, Generic[DcOrArrayT]):
       # Nested dataclass case (if all attributes are `object`, so no active
       # array fields)
       return True
+    elif enp.array_spec.is_fake_array(self.value):
+      # `etree.spec_like`, Flax summary, ShapeDtypeStruct, ... compatibility
+      return True
     return False
 
   @property
